(*
    ===========================================================================
    http://www.na.cse.nagoya-u.ac.jp/~reiji/lect/alg00/index.html
    C ƃASY(Introduction to C Programming)
    ---------------------------------------------------------------------------
    http://www.na.cse.nagoya-u.ac.jp/~reiji/lect/alg00/sec12-1.html
    12.1. ̕\
    ---------------------------------------------------------------------------
    http://www.na.cse.nagoya-u.ac.jp/~reiji/lect/alg00/sec12-2.html
    12.2. 
    ---------------------------------------------------------------------------
    http://www.na.cse.nagoya-u.ac.jp/~reiji/lect/alg00/sec12-3.html
    12.3. \
    ===========================================================================
*)

unit iExpr;

interface

uses
    SysUtils, Classes, Math, fjutil;

type
    //  g[N̎
    TTokenType = (
        ttVAR,          //  ϐ
        ttNUM,          //  萔
        ttLPAR,         //  
        ttRPAR,         //  E
        ttMSIG,         //  
        ttMUL,          //  
        ttDIV,          //  
        ttADD,          //  a
        ttSUB,          //  
        ttPOWER,        //  ݏ    [2004/04/15]
        ttROOT          //  [g  [2004/04/15]
    );

    //  g[N̎ނ
    TTokenTypes = set of TTokenType;

    //  \؂̂ꂼ̒_\f[^\
    TNode = class
    private
        FType       : TTokenType;   //  ̒_\̂̎
        FLeft       : TNode;        //  ̒_Zq̏ꍇÃIyh
        FRight      : TNode;        //  ̏ꍇAFLeft ̕ɃIyh
        FName       : string;       //  ̒_\̂̕\

        //  ZZłΊʂAłȂ΂̂܂܁A
        function    PrintPar1exp(): string;
        //  揜Złʂ
        function    PrintPar2exp(): string;
    public
        //  RXgN^
        constructor Create();
        //  fXgN^
        destructor Destroy(); override;
        //  ʏ̕@(uL@)ň
        function    PrintExp(): string;
        //  uL@(|[hL@)ň
        function    PrintPolish(): string;
        //  vZ
        function    Execute(VarList: TStrings): double;
    end;

    //  ͊
    TTokenizer = class
    private
        FPos        : integer;      //  ǂݎ茻݈ʒu
        FChr        : Char;         //  ݈ʒu̕
        FStr        : string;       //  ǂݎΏە
        FUnGot      : TNode;        //  ǂ݃obt@

        //  i([܂ŗfalse)
        function    IncPos(): Boolean;
        //  ^ǂݗp߂
        procedure   UngetToken(T: TNode);
        //  g[N擾
        function    GetToken(): TNode;
    public
        //  RXgN^
        constructor Create(const S: string);
        //  q
        function    GetFactor(): TNode;
        //  
        function    GetTerm(): TNode;
        //  
        function    GetExp(): TNode;
    end;

const
    //  g[N̎ނ̕\
    TOKEN_STR: array[TTokenType] of string =
    (
        '?', '?', '(', ')', '_', '*', '/', '+', '-', '^', '\'
    );

implementation

(*
    ===========================================================================
    TNode
    ===========================================================================
*)

//-----------------------------------------------------------------------------
//  [2004/04/12]
constructor TNode.Create();
begin
    inherited;
    FType := ttVAR;
    FName := TOKEN_STR[FType];
    FLeft := nil;
    FRight := nil;
end;

//-----------------------------------------------------------------------------
//  [2004/04/15]
destructor TNode.Destroy();
begin
    if (FLeft <> nil) then FLeft.Free;
    if (FRight <> nil) then FRight.Free;
    inherited;
end;

//-----------------------------------------------------------------------------
//  [2004/04/12]
const PAR1_TOKENS :TTokenTypes = [ttADD, ttSUB];
const PAR2_TOKENS :TTokenTypes = [ttADD, ttSUB, ttMUL, ttDIV, ttPOWER];

function    TNode.PrintPar1exp(): string;
begin
    Result := '';
    if (FType in PAR1_TOKENS) then Result := Result + TOKEN_STR[ttLPAR];
    Result := Result + PrintExp();
    if (FType in PAR1_TOKENS) then Result := Result + TOKEN_STR[ttRPAR];
end;
//-----------------------------------------------------------------------------
//  [2004/04/12]
function    TNode.PrintPar2exp(): string;
begin
    Result := '';
    if (FType in PAR2_TOKENS) then Result := Result + TOKEN_STR[ttLPAR];
    Result := Result + PrintExp();
    if (FType in PAR2_TOKENS) then Result := Result + TOKEN_STR[ttRPAR];
end;
//-----------------------------------------------------------------------------
//  [2004/04/12]
function    TNode.PrintExp(): string;
begin
    case FType of

    ttVAR, ttNUM:
        Result := FName;

    ttMSIG, ttROOT:
        Result := TOKEN_STR[FType] + FLeft.PrintPar2exp();

    ttMUL, ttDIV, ttPOWER:
        Result := FLeft.PrintPar1exp()
                + TOKEN_STR[FType]
                + FRight.PrintPar2exp();

    ttADD, ttSUB:
        Result := FLeft.PrintExp()
                + TOKEN_STR[FType]
                + FRight.PrintPar1exp();

    else
        Exception.Create('TNode.PrintExp(): unknown type: ' + TOKEN_STR[FType]);

    end;
end;

//-----------------------------------------------------------------------------
//  [2004/04/12]
function    TNode.PrintPolish(): string;
begin
    case FType of

    ttVAR, ttNUM:
        Result := FName;

    ttMSIG, ttROOT:
        Result := FLeft.PrintPolish()
                + TOKEN_STR[FType];

    ttMUL, ttDIV, ttPOWER, ttADD, ttSUB:
        Result := FLeft.PrintPolish()
                + FRight.PrintPolish()
                + TOKEN_STR[FType];

    else
        Exception.Create('TNode.PrintPolish(): unknown type: ' + TOKEN_STR[FType]);

    end;
end;

//-----------------------------------------------------------------------------
//  [2004/04/15]
function    TNode.Execute(VarList: TStrings): double;
begin
    case FType of

    ttVAR:
        Result := StrToIntDef(VarList.Values[FName], 0);

    ttNUM:
        Result := StrToFloat(FName);

    ttMSIG:
        Result := - FLeft.Execute(VarList);

    ttROOT:
        Result := Power(FLeft.Execute(VarList), 0.5);

    ttMUL:
        Result := FLeft.Execute(VarList) * FRight.Execute(VarList);

    ttDIV:
        try
            Result := FLeft.Execute(VarList) / FRight.Execute(VarList);
        except
            on E:Exception do raise;
        end;

    ttPOWER:
        Result := Power(FLeft.Execute(VarList), FRight.Execute(VarList));

    ttADD:
        Result := FLeft.Execute(VarList) + FRight.Execute(VarList);

    ttSUB:
        Result := FLeft.Execute(VarList) - FRight.Execute(VarList);

    else
        begin
            Exception.Create('TNode.Execute(): unknown type: ' + TOKEN_STR[FType]);
            Result := MaxDouble;
        end;

    end;
end;

(*
    ===========================================================================
    TTokenizer
    ===========================================================================
*)

//-----------------------------------------------------------------------------
constructor TTokenizer.Create(const S: string);
begin
    FPos := 0;
    FStr := S;
    FUnGot := nil;
end;
//-----------------------------------------------------------------------------
function    TTokenizer.IncPos(): Boolean;
begin
    Inc(FPos);
    Result := (FPos <= Length(FStr));
    if (Result) then FChr := FStr[FPos];  
end;
//-----------------------------------------------------------------------------
procedure   TTokenizer.UngetToken(T: TNode);
begin
    FUnGot := T;
end;
//-----------------------------------------------------------------------------
function    TTokenizer.GetToken(): TNode;
begin
    Result := nil;

    //  ݌ɂ΂Ԃ
    if (FUnGot <> nil) then
    begin
        Result := FUnGot;
        FUnGot := nil;
        Exit;
    end;

    //  ݈ʒui߂
    if (not IncPos()) then Exit;
    //   󔒓ǂݔ΂
    while (Checker.isspace(FChr)) do
    begin
        //  ݈ʒui߂
        if (not IncPos()) then Exit;
    end;

    //  _쐬
    Result := TNode.Create();

    //  +
    if (FChr = TOKEN_STR[ttADD]) then
    begin
        Result.FType := ttADD;
        Result.FName := FChr;
    end
    //  -
    else
    if (FChr = TOKEN_STR[ttSUB]) then
    begin
        Result.FType := ttSUB;
        Result.FName := FChr;
    end
    //  -()
    else
    if (FChr = TOKEN_STR[ttMSIG]) then
    begin
        Result.FType := ttMSIG;
        Result.FName := FChr;
    end
    //  *
    else
    if (FChr = TOKEN_STR[ttMUL]) then
    begin
        Result.FType := ttMUL;
        Result.FName := FChr;
    end
    //  /
    else
    if (FChr = TOKEN_STR[ttDIV]) then
    begin
        Result.FType := ttDIV;
        Result.FName := FChr;
    end
    //  ^
    else
    if (FChr = TOKEN_STR[ttPOWER]) then
    begin
        Result.FType := ttPOWER;
        Result.FName := FChr;
    end
    //  \
    else
    if (FChr = TOKEN_STR[ttROOT]) then
    begin
        Result.FType := ttROOT;
        Result.FName := FChr;
    end
    //  (
    else
    if (FChr = TOKEN_STR[ttLPAR]) then
    begin
        Result.FType := ttLPAR;
        Result.FName := FChr;
    end
    //  )
    else
    if (FChr = TOKEN_STR[ttRPAR]) then
    begin
        Result.FType := ttRPAR;
        Result.FName := FChr;
    end
    //  l
    else
    if (Checker.isdigit(FChr) or (FChr = '.')) then
    begin
        Result.FType := ttNUM;
        Result.FName := FChr;

        //  ݈ʒui߂
        if (not IncPos()) then Exit;
        //  ʎqI̊ԃ[v
        while (Checker.isdigit(FChr) or (FChr = '.')) do
        begin
            Result.FName := Result.FName + FChr;
            //  ݈ʒui߂
            if (not IncPos()) then Exit;
        end;
        //  łĂɂ͈i݂
        Dec(FPos);
    end
    //  ʎq
    else
    if (Checker.isalpha(FChr) or (FChr = '_')) then
    begin
        Result.FType := ttVAR;
        Result.FName := FChr;
        //  ݈ʒui߂
        if (not IncPos()) then Exit;
        //  ʎqI̊ԃ[v
        while (Checker.isalnum(FChr) or (FChr = '_')) do
        begin
            Result.FName := Result.FName + FChr;
            //  ݈ʒui߂
            if (not IncPos()) then Exit;
        end;
        //  łĂɂ͈i݂
        Dec(FPos);
    end
    //  Ȃ񂾂ÓH
    else
    begin
        Result.Free;
        Result := nil;
        Exception.Create('TTokenizer.GetToken(): unknown char: ' + FChr);
    end;
end;

//-----------------------------------------------------------------------------
//  [2004/04/13]q
//  factor := literal | ( exp ) | - factor
function    TTokenizer.GetFactor(): TNode;
begin
    Result := GetToken();
    if (Result = nil) then Exit;

    if (Result.FType = ttLPAR) then
    begin
        Result.Free;
        Result := GetExp();
        GetToken().Free;
    end
    else
    if (Result.FType = ttSUB) then
    begin
        Result.FType := ttMSIG;
        Result.FLeft := GetFactor();
    end
    else
    if (Result.FType = ttROOT) then
    begin
        Result.FLeft := GetFactor();
    end;
end;
//-----------------------------------------------------------------------------
//  [2004/04/13]
//  term := factor | term * factor | term / factor
function    TTokenizer.GetTerm(): TNode;
var
    Left, Op: TNode;
begin
    Left := GetFactor();

    while true do
    begin
        Op := GetToken();
        if (Op = nil) then
        begin
            Result := Left;
            Exit;
        end;

        if not (Op.FType in [ttMUL, ttDIV, ttPOWER]) then
        begin
            UngetToken(Op);
            Result := Left;
            Exit;
        end;

        Op.FLeft := Left;
        Op.FRight := GetFactor();
        Left := Op;
    end;

end;
//-----------------------------------------------------------------------------
//  [2004/04/13]
//  exp := term | exp + term | exp - term 
function    TTokenizer.GetExp(): TNode;
var
    Left, Op: TNode;
begin
    Left := GetTerm();
    while true do
    begin
        Op := GetToken();
        if (Op = nil) then
        begin
            Result := Left;
            Exit;
        end;

        if not (Op.FType in [ttADD, ttSUB]) then
        begin
            UngetToken(Op);
            Result := Left;
            Exit;
        end;

        Op.FLeft := Left;
        Op.FRight := GetTerm();
        Left := Op;
    end;
end;

end.

